# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from typing import Any, Dict, List, Union

from supertokens_python.exceptions import raise_bad_input_exception
from supertokens_python.recipe.emailpassword.constants import FORM_FIELD_EMAIL_ID
from supertokens_python.recipe.emailpassword.exceptions import (
    raise_form_field_exception,
)
from supertokens_python.recipe.emailpassword.types import (
    ErrorFormField,
    FormField,
    NormalisedFormField,
)
from supertokens_python.utils import find_first_occurrence_in_list


async def validate_form_or_throw_error(
    inputs: List[FormField], config_form_fields: List[NormalisedFormField]
):
    validation_errors: List[ErrorFormField] = []
    if len(config_form_fields) != len(inputs):
        raise_bad_input_exception("Are you sending too many / too few formFields?")

    for field in config_form_fields:
        input_field: Union[None, FormField] = find_first_occurrence_in_list(
            lambda x: x.id == field.id, inputs
        )
        if input_field is None or (input_field.value == "" and not field.optional):
            validation_errors.append(ErrorFormField(field.id, "Field is not optional"))
        else:
            error = await field.validate(input_field.value)
            if error is not None:
                validation_errors.append(ErrorFormField(field.id, error))

    if len(validation_errors) != 0:
        # raise BadInputError(msg="Error in input formFields")
        raise_form_field_exception("Error in input formFields", validation_errors)


async def validate_form_fields_or_throw_error(
    config_form_fields: List[NormalisedFormField], form_fields_raw: Any
) -> List[FormField]:
    if form_fields_raw is None:
        raise_bad_input_exception("Missing input param: formFields")

    if not isinstance(form_fields_raw, List):
        raise_bad_input_exception("formFields must be an array")

    form_fields: List[FormField] = []

    form_fields_list_raw: List[Dict[str, Any]] = form_fields_raw
    for current_form_field in form_fields_list_raw:
        if (
            "id" not in current_form_field
            or not isinstance(current_form_field["id"], str)
            or "value" not in current_form_field
        ):
            raise_bad_input_exception(
                "All elements of formFields must contain an 'id' and 'value' field"
            )
        value = current_form_field["value"]
        if current_form_field["id"] == FORM_FIELD_EMAIL_ID and isinstance(value, str):
            value = value.strip()
        form_fields.append(FormField(current_form_field["id"], value))

    await validate_form_or_throw_error(form_fields, config_form_fields)
    return form_fields



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from typing import TYPE_CHECKING, Callable, Union

from supertokens_python.ingredients.emaildelivery.types import EmailDeliveryConfig
from supertokens_python.recipe.emailpassword.types import EmailTemplateVars

from . import exceptions as ex
from . import utils
from .emaildelivery import services as emaildelivery_services
from .recipe import EmailPasswordRecipe

exceptions = ex
InputOverrideConfig = utils.InputOverrideConfig
InputResetPasswordUsingTokenFeature = utils.InputResetPasswordUsingTokenFeature
InputSignUpFeature = utils.InputSignUpFeature
InputFormField = utils.InputFormField
SMTPService = emaildelivery_services.SMTPService

if TYPE_CHECKING:
    from supertokens_python.supertokens import AppInfo

    from ...recipe_module import RecipeModule


def init(
    sign_up_feature: Union[utils.InputSignUpFeature, None] = None,
    reset_password_using_token_feature: Union[
        utils.InputResetPasswordUsingTokenFeature, None
    ] = None,
    override: Union[utils.InputOverrideConfig, None] = None,
    email_delivery: Union[EmailDeliveryConfig[EmailTemplateVars], None] = None,
) -> Callable[[AppInfo], RecipeModule]:
    return EmailPasswordRecipe.init(
        sign_up_feature,
        reset_password_using_token_feature,
        override,
        email_delivery,
    )



from typing import Any, Dict, List, Union

from supertokens_python.exceptions import raise_bad_input_exception
from supertokens_python.recipe.session.asyncio import revoke_multiple_sessions
from ...interfaces import APIInterface, APIOptions, UserSessionsPostAPIResponse


async def handle_user_sessions_post(
    _api_interface: APIInterface, api_options: APIOptions
) -> UserSessionsPostAPIResponse:
    request_body: Dict[str, Any] = await api_options.request.json()  # type: ignore
    session_handles: Union[List[str], Any] = request_body.get("sessionHandles")

    if session_handles is None or not isinstance(session_handles, list):
        return raise_bad_input_exception(
            "Required parameter 'sessionHandles' is missing or has an invalid type"
        )

    await revoke_multiple_sessions(session_handles)
    return UserSessionsPostAPIResponse()



from typing import Any, Dict, Union

from supertokens_python.exceptions import raise_bad_input_exception
from supertokens_python.recipe.emailverification.asyncio import (
    create_email_verification_token,
    send_email,
)
from supertokens_python.recipe.emailverification.interfaces import (
    CreateEmailVerificationTokenEmailAlreadyVerifiedError,
    CreateEmailVerificationTokenOkResult,
)
from supertokens_python.recipe.emailverification.recipe import (
    EmailVerificationRecipe,
    GetEmailForUserIdOkResult,
)
from supertokens_python.recipe.emailverification.types import (
    VerificationEmailTemplateVars,
    VerificationEmailTemplateVarsUser,
)
from supertokens_python.recipe.emailverification.utils import get_email_verify_link

from ...interfaces import (
    APIInterface,
    APIOptions,
    UserEmailVerifyTokenPostAPIEmailAlreadyVerifiedErrorResponse,
    UserEmailVerifyTokenPostAPIOkResponse,
)


async def handle_email_verify_token_post(
    _api_interface: APIInterface, api_options: APIOptions
) -> Union[
    UserEmailVerifyTokenPostAPIOkResponse,
    UserEmailVerifyTokenPostAPIEmailAlreadyVerifiedErrorResponse,
]:
    request_body: Dict[str, Any] = await api_options.request.json()  # type: ignore
    user_id = request_body.get("userId")

    if user_id is None or not isinstance(user_id, str):
        raise_bad_input_exception(
            "Required parameter 'userId' is missing or has an invalid type"
        )

    email_response = await EmailVerificationRecipe.get_instance().get_email_for_user_id(
        user_id, {}
    )

    if not isinstance(email_response, GetEmailForUserIdOkResult):
        raise Exception("Should not come here")

    email_verification_token = await create_email_verification_token(user_id)

    if isinstance(
        email_verification_token, CreateEmailVerificationTokenEmailAlreadyVerifiedError
    ):
        return UserEmailVerifyTokenPostAPIEmailAlreadyVerifiedErrorResponse()

    assert isinstance(email_verification_token, CreateEmailVerificationTokenOkResult)

    email_verify_link = get_email_verify_link(
        api_options.app_info, email_verification_token.token, user_id
    )

    await send_email(
        VerificationEmailTemplateVars(
            user=VerificationEmailTemplateVarsUser(user_id, email_response.email),
            email_verify_link=email_verify_link,
            user_context={},
        )
    )

    return UserEmailVerifyTokenPostAPIOkResponse()



from typing import Any, Dict, List

from fastapi import FastAPI
from pytest import fixture, mark
from starlette.testclient import TestClient
from supertokens_python import init
from supertokens_python.constants import DASHBOARD_VERSION
from supertokens_python.framework import BaseRequest
from supertokens_python.framework.fastapi import get_middleware
from supertokens_python.recipe import dashboard, emailpassword, session, usermetadata
from supertokens_python.recipe.dashboard import InputOverrideConfig
from supertokens_python.recipe.dashboard.interfaces import (
    RecipeInterface as DashboardRI,
)
from supertokens_python.recipe.dashboard.utils import DashboardConfig
from supertokens_python.recipe.usermetadata.asyncio import update_user_metadata
from tests.utils import (
    clean_st,
    get_st_init_args,
    min_api_version,
    reset,
    setup_st,
    sign_up_request,
    start_st,
)


def setup_function(_):
    reset()
    clean_st()
    setup_st()


def teardown_function(_):
    reset()
    clean_st()


pytestmark = mark.asyncio


@fixture(scope="function")
async def app():
    app = FastAPI()
    app.add_middleware(get_middleware())

    return TestClient(app)


async def test_dashboard_recipe(app: TestClient):
    def override_dashboard_functions(oi: DashboardRI) -> DashboardRI:
        async def should_allow_access(
            _request: BaseRequest,
            _config: DashboardConfig,
            _user_context: Dict[str, Any],
        ) -> bool:
            return True

        oi.should_allow_access = should_allow_access
        return oi

    st_args = get_st_init_args(
        [
            session.init(get_token_transfer_method=lambda _, __, ___: "cookie"),
            dashboard.init(
                api_key="someKey",
                override=InputOverrideConfig(functions=override_dashboard_functions),
            ),
        ]
    )
    init(**st_args)
    start_st()

    expected_url = (
        f"https://cdn.jsdelivr.net/gh/supertokens/dashboard@v{DASHBOARD_VERSION}/build/"
    )

    res = app.get(url="/auth/dashboard")
    assert res.status_code == 200
    assert expected_url in str(res.text)


@min_api_version("2.13")
async def test_dashboard_users_get(app: TestClient):
    def override_dashboard_functions(oi: DashboardRI) -> DashboardRI:
        async def should_allow_access(
            _request: BaseRequest,
            _config: DashboardConfig,
            _user_context: Dict[str, Any],
        ) -> bool:
            return True

        oi.should_allow_access = should_allow_access
        return oi

    st_args = get_st_init_args(
        [
            session.init(get_token_transfer_method=lambda _, __, ___: "cookie"),
            emailpassword.init(),
            usermetadata.init(),
            dashboard.init(
                api_key="someKey",
                override=InputOverrideConfig(
                    functions=override_dashboard_functions,
                ),
            ),
        ]
    )
    init(**st_args)
    start_st()

    user_ids: List[str] = []

    # Create two emailpassword users:
    for i in range(2):
        res = sign_up_request(app, f"user{i}@example.com", "password123")
        user_id: str = res.json()["user"]["id"]
        user_ids.append(user_id)
        assert res.status_code == 200

    await update_user_metadata(user_ids[0], {"first_name": "User1", "last_name": "Foo"})
    await update_user_metadata(user_ids[1], {"first_name": "User2"})

    res = app.get(url="/auth/dashboard/api/users?limit=5")
    body = res.json()
    assert res.status_code == 200
    assert body["users"][0]["user"]["firstName"] == "User2"
    assert body["users"][1]["user"]["lastName"] == "Foo"
    assert body["users"][1]["user"]["firstName"] == "User1"



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from os import environ
from typing import TYPE_CHECKING, Any, Dict, List, Union

from supertokens_python.framework.response import BaseResponse
from supertokens_python.ingredients.emaildelivery.types import EmailDeliveryConfig
from supertokens_python.normalised_url_path import NormalisedURLPath
from supertokens_python.querier import Querier
from supertokens_python.recipe.passwordless.types import PasswordlessIngredients
from supertokens_python.recipe.thirdparty.provider import Provider
from supertokens_python.recipe.thirdparty.types import ThirdPartyIngredients
from supertokens_python.recipe.thirdpartypasswordless.types import (
    ThirdPartyPasswordlessIngredients,
)
from supertokens_python.recipe_module import APIHandled, RecipeModule

from ..passwordless.utils import ContactConfig
from .api.implementation import APIImplementation
from .api.passwordless_api_impementation import (
    get_interface_impl as get_passwordless_interface_impl,
)
from .api.thirdparty_api_implementation import (
    get_interface_impl as get_third_party_interface_impl,
)
from .recipeimplementation.implementation import RecipeImplementation
from .recipeimplementation.passwordless_recipe_implementation import (
    RecipeImplementation as PasswordlessRecipeImplementation,
)
from .recipeimplementation.third_party_recipe_implementation import (
    RecipeImplementation as ThirdPartyRecipeImplementation,
)

if TYPE_CHECKING:
    from supertokens_python.framework.request import BaseRequest
    from supertokens_python.supertokens import AppInfo

from typing import Awaitable, Callable

from supertokens_python.exceptions import SuperTokensError
from supertokens_python.ingredients.emaildelivery import EmailDeliveryIngredient
from supertokens_python.ingredients.smsdelivery import SMSDeliveryIngredient
from supertokens_python.ingredients.smsdelivery.types import SMSDeliveryConfig
from supertokens_python.recipe.thirdparty import ThirdPartyRecipe
from supertokens_python.recipe.thirdparty.utils import (
    InputOverrideConfig as TPOverrideConfig,
)
from supertokens_python.recipe.thirdparty.utils import SignInAndUpFeature
from typing_extensions import Literal

from ..passwordless import PasswordlessRecipe
from ..passwordless.interfaces import APIInterface as PasswordlessAPIInterface
from ..passwordless.interfaces import RecipeInterface as PasswordlessRecipeInterface
from ..passwordless.utils import OverrideConfig as PlessOverrideConfig
from ..thirdparty.interfaces import APIInterface as ThirdPartyAPIInterface
from ..thirdparty.interfaces import RecipeInterface as ThirdPartyRecipeInterface
from .exceptions import SupertokensThirdPartyPasswordlessError
from .interfaces import APIInterface, RecipeInterface
from .types import EmailTemplateVars, SMSTemplateVars
from .utils import InputOverrideConfig, validate_and_normalise_user_input


class ThirdPartyPasswordlessRecipe(RecipeModule):
    recipe_id = "thirdpartypasswordless"
    __instance = None
    email_delivery: EmailDeliveryIngredient[EmailTemplateVars]
    sms_delivery: SMSDeliveryIngredient[SMSTemplateVars]

    def __init__(
        self,
        recipe_id: str,
        app_info: AppInfo,
        contact_config: ContactConfig,
        flow_type: Literal[
            "USER_INPUT_CODE", "MAGIC_LINK", "USER_INPUT_CODE_AND_MAGIC_LINK"
        ],
        ingredients: ThirdPartyPasswordlessIngredients,
        get_custom_user_input_code: Union[
            Callable[[Dict[str, Any]], Awaitable[str]], None
        ] = None,
        override: Union[InputOverrideConfig, None] = None,
        providers: Union[List[Provider], None] = None,
        third_party_recipe: Union[ThirdPartyRecipe, None] = None,
        passwordless_recipe: Union[PasswordlessRecipe, None] = None,
        email_delivery: Union[EmailDeliveryConfig[EmailTemplateVars], None] = None,
        sms_delivery: Union[SMSDeliveryConfig[SMSTemplateVars], None] = None,
    ):
        super().__init__(recipe_id, app_info)
        self.config = validate_and_normalise_user_input(
            self,
            contact_config=contact_config,
            flow_type=flow_type,
            get_custom_user_input_code=get_custom_user_input_code,
            override=override,
            providers=providers,
            email_delivery=email_delivery,
            sms_delivery=sms_delivery,
        )

        recipe_implementation = RecipeImplementation(
            Querier.get_instance(PasswordlessRecipe.recipe_id),
            Querier.get_instance(ThirdPartyRecipe.recipe_id),
        )
        self.recipe_implementation: RecipeInterface = (
            recipe_implementation
            if self.config.override.functions is None
            else self.config.override.functions(recipe_implementation)
        )

        api_implementation = APIImplementation()
        self.api_implementation: APIInterface = (
            api_implementation
            if self.config.override.apis is None
            else self.config.override.apis(api_implementation)
        )

        email_delivery_ingredient = ingredients.email_delivery
        if email_delivery_ingredient is None:
            self.email_delivery = EmailDeliveryIngredient(
                self.config.get_email_delivery_config()
            )
        else:
            self.email_delivery = email_delivery_ingredient

        sms_delivery_ingredient = ingredients.sms_delivery
        if sms_delivery_ingredient is None:
            self.sms_delivery = SMSDeliveryIngredient(
                self.config.get_sms_delivery_config()
            )
        else:
            self.sms_delivery = sms_delivery_ingredient

        if passwordless_recipe is not None:
            self.passwordless_recipe = passwordless_recipe
        else:

            def func_override_passwordless(
                _: PasswordlessRecipeInterface,
            ) -> PasswordlessRecipeInterface:
                return PasswordlessRecipeImplementation(self.recipe_implementation)

            def apis_override_passwordless(
                _: PasswordlessAPIInterface,
            ) -> PasswordlessAPIInterface:
                return get_passwordless_interface_impl(self.api_implementation)

            pless_email_delivery = self.email_delivery
            pless_sms_delivery = self.sms_delivery
            pless_ingredients = PasswordlessIngredients(
                pless_email_delivery, pless_sms_delivery
            )
            self.passwordless_recipe = PasswordlessRecipe(
                recipe_id,
                app_info,
                self.config.contact_config,
                self.config.flow_type,
                pless_ingredients,
                PlessOverrideConfig(
                    func_override_passwordless, apis_override_passwordless
                ),
                self.config.get_custom_user_input_code,
            )

        if third_party_recipe is not None:
            self.third_party_recipe: Union[ThirdPartyRecipe, None] = third_party_recipe
        else:

            def func_override_third_party(
                _: ThirdPartyRecipeInterface,
            ) -> ThirdPartyRecipeInterface:
                return ThirdPartyRecipeImplementation(self.recipe_implementation)

            def apis_override_third_party(
                _: ThirdPartyAPIInterface,
            ) -> ThirdPartyAPIInterface:
                return get_third_party_interface_impl(self.api_implementation)

            self.third_party_recipe: Union[ThirdPartyRecipe, None] = None

            if len(self.config.providers) != 0:
                tp_ingredients = ThirdPartyIngredients()
                self.third_party_recipe = ThirdPartyRecipe(
                    recipe_id,
                    app_info,
                    SignInAndUpFeature(self.config.providers),
                    tp_ingredients,
                    TPOverrideConfig(
                        func_override_third_party, apis_override_third_party
                    ),
                )

    def is_error_from_this_recipe_based_on_instance(self, err: Exception) -> bool:
        return isinstance(err, SuperTokensError) and (
            isinstance(err, SupertokensThirdPartyPasswordlessError)
            or self.passwordless_recipe.is_error_from_this_recipe_based_on_instance(err)
            or (
                self.third_party_recipe is not None
                and self.third_party_recipe.is_error_from_this_recipe_based_on_instance(
                    err
                )
            )
        )

    def get_apis_handled(self) -> List[APIHandled]:
        apis_handled = self.passwordless_recipe.get_apis_handled()
        if self.third_party_recipe is not None:
            apis_handled = apis_handled + self.third_party_recipe.get_apis_handled()
        return apis_handled

    async def handle_api_request(
        self,
        request_id: str,
        request: BaseRequest,
        path: NormalisedURLPath,
        method: str,
        response: BaseResponse,
    ):
        if (
            self.passwordless_recipe.return_api_id_if_can_handle_request(path, method)
            is not None
        ):
            return await self.passwordless_recipe.handle_api_request(
                request_id, request, path, method, response
            )
        if (
            self.third_party_recipe is not None
            and self.third_party_recipe.return_api_id_if_can_handle_request(
                path, method
            )
            is not None
        ):
            return await self.third_party_recipe.handle_api_request(
                request_id, request, path, method, response
            )
        return None

    async def handle_error(
        self, request: BaseRequest, err: SuperTokensError, response: BaseResponse
    ) -> BaseResponse:
        if self.passwordless_recipe.is_error_from_this_recipe_based_on_instance(err):
            return await self.passwordless_recipe.handle_error(request, err, response)
        if (
            self.third_party_recipe is not None
            and self.third_party_recipe.is_error_from_this_recipe_based_on_instance(err)
        ):
            return await self.third_party_recipe.handle_error(request, err, response)
        raise err

    def get_all_cors_headers(self) -> List[str]:
        cors_headers = self.passwordless_recipe.get_all_cors_headers()
        if self.third_party_recipe is not None:
            cors_headers = cors_headers + self.third_party_recipe.get_all_cors_headers()
        return cors_headers

    @staticmethod
    def init(
        contact_config: ContactConfig,
        flow_type: Literal[
            "USER_INPUT_CODE", "MAGIC_LINK", "USER_INPUT_CODE_AND_MAGIC_LINK"
        ],
        get_custom_user_input_code: Union[
            Callable[[Dict[str, Any]], Awaitable[str]], None
        ] = None,
        email_delivery: Union[EmailDeliveryConfig[EmailTemplateVars], None] = None,
        sms_delivery: Union[SMSDeliveryConfig[SMSTemplateVars], None] = None,
        override: Union[InputOverrideConfig, None] = None,
        providers: Union[List[Provider], None] = None,
    ):
        def func(app_info: AppInfo):
            if ThirdPartyPasswordlessRecipe.__instance is None:
                ingredients = ThirdPartyPasswordlessIngredients(None, None)
                ThirdPartyPasswordlessRecipe.__instance = ThirdPartyPasswordlessRecipe(
                    ThirdPartyPasswordlessRecipe.recipe_id,
                    app_info,
                    contact_config,
                    flow_type,
                    ingredients,
                    get_custom_user_input_code,
                    override,
                    providers,
                    email_delivery=email_delivery,
                    sms_delivery=sms_delivery,
                )
                return ThirdPartyPasswordlessRecipe.__instance
            raise Exception(
                None,
                "ThirdPartyPasswordlessRecipe recipe has already been initialised. Please check your code for bugs.",
            )

        return func

    @staticmethod
    def get_instance() -> ThirdPartyPasswordlessRecipe:
        if ThirdPartyPasswordlessRecipe.__instance is not None:
            return ThirdPartyPasswordlessRecipe.__instance
        raise Exception(
            None,
            "Initialisation not done. Did you forget to call the SuperTokens.init function?",
        )

    @staticmethod
    def reset():
        if ("SUPERTOKENS_ENV" not in environ) or (
            environ["SUPERTOKENS_ENV"] != "testing"
        ):
            raise Exception(None, "calling testing function in non testing env")
        ThirdPartyPasswordlessRecipe.__instance = None



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from typing import Any, Dict
from supertokens_python.recipe.jwt.interfaces import CreateJwtResultUnsupportedAlgorithm

from supertokens_python.recipe.openid.interfaces import RecipeInterface

from .constants import ACCESS_TOKEN_PAYLOAD_JWT_PROPERTY_NAME_KEY


async def add_jwt_to_access_token_payload(
    access_token_payload: Dict[str, Any],
    jwt_expiry: int,
    user_id: str,
    jwt_property_name: str,
    openid_recipe_implementation: RecipeInterface,
    user_context: Dict[str, Any],
):
    if ACCESS_TOKEN_PAYLOAD_JWT_PROPERTY_NAME_KEY in access_token_payload:
        # If jwtPropertyName is not undefined it means that the JWT was added
        # to the access token payload already
        existing_jwt_property_name = access_token_payload[
            ACCESS_TOKEN_PAYLOAD_JWT_PROPERTY_NAME_KEY
        ]

        # Delete the old JWT and the old property name
        if existing_jwt_property_name in access_token_payload:
            del access_token_payload[existing_jwt_property_name]
        del access_token_payload[ACCESS_TOKEN_PAYLOAD_JWT_PROPERTY_NAME_KEY]

    # Create the JWT
    jwt_response = await openid_recipe_implementation.create_jwt(
        {
            # We add our claims before the user provided ones so that if they use the same claims
            # then the final payload will use the values they provide
            "sub": user_id,
            **access_token_payload,
        },
        jwt_expiry,
        user_context,
    )

    if isinstance(jwt_response, CreateJwtResultUnsupportedAlgorithm):
        # Should never come here
        raise Exception("JWT Signing algorithm not supported")

    # Add the jwt and the property name to the access token payload
    # We add the JWT after the user defined keys because we want to make sure that it never
    # gets overwritten by a user defined key. Using the same key as the one configured (or defaulting)
    # for the JWT should be considered a dev error
    #
    # ACCESS_TOKEN_PAYLOAD_JWT_PROPERTY_NAME_KEY indicates a reserved key used to determine the property name
    # with which the JWT is set, used to retrieve the JWT from the access token payload during refresh and
    # updateAccessTokenPayload
    #
    # Note: If the user has multiple overrides each with a unique propertyNameInAccessTokenPayload, the logic for
    # checking the existing JWT when refreshing the session or updating the access token payload will not work.
    # This is because even though the jwt itself would be created with unique property names, the _jwtPName value
    # would always be overwritten by the override that runs last and when retrieving the jwt using that key name
    # it cannot be guaranteed that the right JWT is returned. This case is considered to be a rare requirement
    # and we assume that users will not need multiple JWT representations of
    # their access token payload.
    access_token_payload[jwt_property_name] = jwt_response.jwt
    access_token_payload[ACCESS_TOKEN_PAYLOAD_JWT_PROPERTY_NAME_KEY] = jwt_property_name

    return access_token_payload



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from abc import ABC, abstractmethod
from typing import Any, Dict, Optional


class BaseResponse(ABC):
    @abstractmethod
    def __init__(self, content: Dict[str, Any], status_code: int = 200):
        self.content = content
        self.status_code = status_code
        self.wrapper_used = True

    @abstractmethod
    def set_cookie(
        self,
        key: str,
        value: str,
        #    max_age: Union[int, None] = None,
        expires: int,
        path: str = "/",
        domain: Optional[str] = None,
        secure: bool = False,
        httponly: bool = False,
        samesite: str = "lax",
    ):
        pass

    @abstractmethod
    def set_header(self, key: str, value: str) -> None:
        pass

    @abstractmethod
    def get_header(self, key: str) -> Optional[str]:
        pass

    @abstractmethod
    def remove_header(self, key: str) -> None:
        pass

    @abstractmethod
    def set_status_code(self, status_code: int):
        pass

    @abstractmethod
    def set_json_content(self, content: Dict[str, Any]):
        pass

    @abstractmethod
    def set_html_content(self, content: str):
        pass



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from os import environ
from typing import Any, Awaitable, Callable, Dict, Union

from httpx import AsyncClient
from supertokens_python.ingredients.emaildelivery.types import EmailDeliveryInterface
from supertokens_python.logger import log_debug_message
from supertokens_python.recipe.emailverification.types import (
    User,
    VerificationEmailTemplateVars,
)
from supertokens_python.supertokens import AppInfo
from supertokens_python.utils import handle_httpx_client_exceptions


def default_create_and_send_custom_email(
    app_info: AppInfo,
) -> Callable[[User, str, Dict[str, Any]], Awaitable[None]]:
    async def func(user: User, email_verification_url: str, _: Dict[str, Any]):
        if ("SUPERTOKENS_ENV" in environ) and (environ["SUPERTOKENS_ENV"] == "testing"):
            return
        data = {
            "email": user.email,
            "appName": app_info.app_name,
            "emailVerifyURL": email_verification_url,
        }
        try:
            async with AsyncClient() as client:
                resp = await client.post("https://api.supertokens.io/0/st/auth/email/verify", json=data, headers={"api-version": "0"})  # type: ignore
                resp.raise_for_status()
                log_debug_message("Email verification email sent to %s", user.email)
        except Exception as e:
            log_debug_message("Error sending verification email")
            handle_httpx_client_exceptions(e, data)

    return func


class BackwardCompatibilityService(
    EmailDeliveryInterface[VerificationEmailTemplateVars]
):
    def __init__(
        self,
        app_info: AppInfo,
        create_and_send_custom_email: Union[
            Callable[[User, str, Dict[str, Any]], Awaitable[None]], None
        ] = None,
    ) -> None:
        self.create_and_send_custom_email = (
            default_create_and_send_custom_email(app_info)
            if create_and_send_custom_email is None
            else create_and_send_custom_email
        )

    async def send_email(
        self, template_vars: VerificationEmailTemplateVars, user_context: Dict[str, Any]
    ) -> None:
        try:
            email_user = User(template_vars.user.id, template_vars.user.email)
            await self.create_and_send_custom_email(
                email_user, template_vars.email_verify_link, user_context
            )
        except Exception as _:
            pass
